"""SWAG exception classes"""


class InvalidSWAGDataException(Exception):
    pass


class SWAGException(Exception):
    pass


class InvalidSWAGBackendError(SWAGException, ImportError):
    pass


class MissingSWAGParameter(SWAGException):
    pass



import logging

import boto3
import jmespath

from swag_client.backend import SWAGManager

logger = logging.getLogger(__file__)


class DynamoDBSWAGManager(SWAGManager):
    def __init__(self, namespace, **kwargs):
        """Create a DynamoDb based SWAG backend."""
        self.namespace = namespace
        resource = boto3.resource('dynamodb', region_name=kwargs['region'])
        self.table = resource.Table(namespace)

    def get(self, search_filter):
        """Fetch one item from file. Applying given filter, raises exception if more than one item found."""
        logger.debug('Fetching items. Filter: {filter}. Table: {namespace}'.format(
            filter=search_filter,
            namespace=self.namespace
        ))

        items = self.table.scan()['Items']
        results = jmespath.search(search_filter, items)

        if len(results) > 1:
            raise Exception('Found more than one result for get!')

        elif len(results) == 1:
            return results[0]

    def create(self, item):
        """Creates a new item in file."""
        logger.debug('Creating new item. Item: {item} Table: {namespace}'.format(
            item=item,
            namespace=self.namespace
        ))

        self.table.put_item(Item=item)

        return item

    def delete(self, item):
        """Deletes item in file."""
        logger.debug('Deleting item. Item: {item} Table: {namespace}'.format(
            item=item,
            namespace=self.namespace
        ))

        self.table.delete_item(Key={'id': item['id']})

    def update(self, item):
        """Updates item info in file."""
        logger.debug('Updating item. Item: {item} Table: {namespace}'.format(
            item=item,
            namespace=self.namespace
        ))
        self.table.put_item(Item=item)
        return item

    def get_all(self, search_filter=None):
        """Gets all items in file."""
        logger.debug('Fetching items. Filter: {filter}. Table: {namespace}'.format(
            filter=search_filter,
            namespace=self.namespace
        ))

        items = self.table.scan()['Items']

        if search_filter:
            return jmespath.search(search_filter, items)

        return items



"""
.. module:: swag_client.backend
    :platform: Unix
.. author:: Kevin Glisson (kglisson@netflix.com)
"""
import logging
import pkg_resources

import jmespath
from beaker.cache import cache_regions, cache_region

from swag_client.schemas import v1, v2
from swag_client.exceptions import InvalidSWAGBackendError, InvalidSWAGDataException

logger = logging.getLogger(__name__)


def validate(item, namespace='accounts', version=2):
    """Validate item against version schema."""
    if namespace == 'accounts':
        if version == 2:
            return v2.AccountSchema(strict=True).load(item).data
        elif version == 1:
            return v1.AccountSchema(strict=True).load(item).data
        raise InvalidSWAGDataException('Schema version is not supported. Version: {}'.format(version))
    raise InvalidSWAGDataException('Namespace not supported. Namespace: {}'.format(namespace))


def one(items):
    """Fetches one item from a list. Throws exception if there are multiple items."""
    if items:
        if len(items) > 1:
            raise InvalidSWAGBackendError('Attempted to fetch one item, but multiple items found.')

        return items[0]


def get(name):
    for ep in pkg_resources.iter_entry_points('swag_client.backends'):
        if ep.name == name:
            return ep.load()


class SWAGManager(object):
    """Manages swag backends."""
    def __init__(self, *args, **kwargs):
        self.version = kwargs['schema_version']
        self.backend = get(kwargs['type'])(*args, **kwargs)
        cache_regions.update({
            'swag': {
                'expire': kwargs['cache_expires'],
                'type': 'memory'
            }
        })

    def create(self, item):
        """Create a new item in backend."""
        return self.backend.create(validate(item, version=self.version))

    def delete(self, item):
        """Delete an item in backend."""
        return self.backend.delete(item)

    def update(self, item):
        """Update an item in backend."""
        return self.backend.update(validate(item, version=self.version))

    @cache_region('swag')
    def get(self, search_filter):
        """Fetch one item from backend."""
        return self.backend.get(search_filter)

    @cache_region('swag')
    def get_all(self, search_filter=None):
        """Fetch all data from backend."""
        return self.backend.get_all(search_filter=search_filter)

    def get_service_enabled(self, service_name, accounts_list=None):
        if not accounts_list:
            accounts_list = self.backend.get_all().get('accounts', [])

        service_enabled_accounts = []

        for account in accounts_list:
            for service in account.get('services', []):
                # The part we need to get to is a little buried.  Basically, we need to scan the list of services for
                # a dictionary where the name key matches, then we have to scan the status key, which is a list, and
                # then look for a dictionary where the enabled key is set to True
                if(service['name'] == service_name and
                   len(list(filter(lambda status: status['enabled'], service.get('status', []))))):
                    service_enabled_accounts.append(account)
                    break
        return service_enabled_accounts

    def get_service(self, name, search_filter):
        """Fetch service metadata."""
        service_filter = "services[?name=='{}']".format(name)
        return one(jmespath.search(service_filter, self.get(search_filter)))

    def get_service_name(self, name, search_filter):
        """Fetch account name as referenced by a particular service. """
        service_filter = "services[?name=='{}'].metadata.name".format(name)
        return one(jmespath.search(service_filter, self.get(search_filter)))

    def get_by_name(self, name, alias=None):
        """Fetch all accounts with name specified, optionally include aliases."""
        search_filter = "[?name=='{}']".format(name)

        if alias:
            if self.version == 1:
                search_filter = "accounts[?name=='{name}' || alias[?contains(@, '{name}') == `true`]]".format(name=name)

            elif self.version == 2:
                search_filter = "[?name=='{name}' || aliases[?contains(@, '{name}') == `true`]]".format(name=name)

        return self.get_all(search_filter)



import os
import time
import json
import click
from tabulate import tabulate

from swag_client.backend import SWAGManager, validate
from swag_client.migrations import run_migration
from swag_client.util import parse_swag_config_options


def create_swag_from_ctx(ctx):
    """Creates SWAG client from the current context."""
    swag_opts = {}
    if ctx.obj['TYPE'] == 'file':
        swag_opts = {
            'swag.type': 'file',
            'swag.data_dir': ctx.obj['DATA_DIR'],
            'swag.data_file': ctx.obj['DATA_FILE']
        }
    elif ctx.obj['TYPE'] == 's3':
        swag_opts = {
            'swag.type': 's3',
            'swag.bucket_name': ctx.obj['BUCKET_NAME'],
            'swag.data_file': ctx.obj['DATA_FILE'],
            'swag.region': ctx.obj['REGION']
        }
    elif ctx.obj['TYPE'] == 'dynamodb':
        swag_opts = {
            'swag.type': 'dynamodb',
            'swag.region': ctx.obj['REGION']
        }
    return SWAGManager(**parse_swag_config_options(swag_opts))


@click.group()
@click.option('--namespace', default='accounts')
@click.pass_context
def cli(ctx, namespace):
    ctx.obj = {'NAMESPACE': namespace}


@cli.group()
@click.option('--region', default='us-east-1', help='Region the table is located in.')
@click.pass_context
def dynamodb(ctx, region):
    ctx.obj['REGION'] = region
    ctx.obj['TYPE'] = 'dynamodb'


@cli.group()
@click.option('--data-dir', help='Directory to store data.', default=os.getcwd())
@click.option('--data-file')
@click.pass_context
def file(ctx, data_dir, data_file):
    """Use the File SWAG Backend"""
    ctx.obj['DATA_DIR'] = data_dir
    ctx.obj['DATA_FILE'] = data_file
    ctx.obj['TYPE'] = 'file'


@cli.group()
@click.option('--bucket-name', help='Name of the bucket you wish to operate on.')
@click.option('--data-file', help='Key name of the file to operate on.')
@click.option('--region', default='us-east-1', help='Region the bucket is located in.')
@click.pass_context
def s3(ctx, bucket_name, data_file, region):
    """Use the S3 SWAG backend."""
    ctx.obj['BUCKET_NAME'] = bucket_name
    ctx.obj['DATA_FILE'] = data_file
    ctx.obj['TYPE'] = 's3'
    ctx.obj['REGION'] = region


@cli.command()
@click.pass_context
def list(ctx):
    """List SWAG account info."""
    if ctx.obj.get('NAMESPACE') != 'accounts':
        click.echo(
            click.style('Only account data is available for listing.', fg='red')
        )
        return

    swag = create_swag_from_ctx(ctx)
    accounts = swag.get_all()
    _table = [[result['name'], result.get('id')] for result in accounts]
    click.echo(
        tabulate(_table, headers=["Account Name", "Account Number"])
    )


@cli.command()
@click.option('--start-version', default=1, help='Starting version.')
@click.option('--end-version', default=2, help='Ending version.')
@click.pass_context
def migrate(ctx, start_version, end_version):
    """Transition from one SWAG schema to another."""
    if ctx.obj['TYPE'] == 'file':
        if ctx.obj['DATA_FILE']:
            file_path = ctx.obj['DATA_FILE']
        else:
            file_path = os.path.join(ctx.obj['DATA_DIR'], ctx.obj['NAMESPACE'] + '.json')

        # todo make this more like alemebic and determine/load versions automatically
        with open(file_path, 'r') as f:
            data = json.loads(f.read())

        data = run_migration(data, start_version, end_version)
        with open(file_path, 'w') as f:
            f.write(json.dumps(data))


@cli.command()
@click.pass_context
def propagate(ctx):
    """Transfers SWAG data from one backend to another"""
    data = []
    if ctx.obj['TYPE'] == 'file':
        if ctx.obj['DATA_FILE']:
            file_path = ctx.obj['DATA_FILE']
        else:
            file_path = os.path.join(ctx.obj['DATA_DIR'], ctx.obj['NAMESPACE'] + '.json')

        with open(file_path, 'r') as f:
            data = json.loads(f.read())

    swag_opts = {
        'swag.type': 'dynamodb'
    }

    swag = SWAGManager(**parse_swag_config_options(swag_opts))

    for item in data:
        time.sleep(2)
        swag.create(item)


@cli.command()
@click.pass_context
def create(ctx):
    """Create a new SWAG item."""
    pass


@cli.command()
@click.pass_context
@click.argument('data', type=click.File())
def update(ctx, data):
    """Updates a given record."""
    swag = create_swag_from_ctx(ctx)
    data = json.loads(data.read())

    for account in data:
        swag.update(account)


@cli.command()
@click.argument('data', type=click.File())
@click.pass_context
def seed_aws_data(ctx, data):
    """Seeds SWAG from a list of known AWS accounts."""
    swag = create_swag_from_ctx(ctx)
    for k, v in json.loads(data.read()).items():
        for account in v['accounts']:
            data = {
                    'description': 'This is an AWS owned account used for {}'.format(k),
                    'id': account['account_id'],
                    'contacts': [],
                    'owner': 'aws',
                    'provider': 'aws',
                    'sensitive': False,
                    'email': 'support@amazon.com',
                    'name': k + '-' + account['region']
                }

            click.echo(click.style(
                'Seeded Account. AccountName: {}'.format(data['name']), fg='green')
            )

            swag.create(data)


file.add_command(list)
file.add_command(validate)
file.add_command(migrate)
file.add_command(propagate)
file.add_command(create)
file.add_command(seed_aws_data)
file.add_command(update)
dynamodb.add_command(list)
dynamodb.add_command(create)
dynamodb.add_command(update)
dynamodb.add_command(seed_aws_data)
s3.add_command(list)
s3.add_command(create)
s3.add_command(update)
s3.add_command(seed_aws_data)





from datetime import datetime
from marshmallow import Schema, fields
from marshmallow.validate import OneOf

PROVIDERS = ['aws', 'gcp', 'azure']
ACCOUNT_TYPES = ['billing', 'security', 'shared-service', 'service']
ACCOUNT_STATUSES = ['created', 'in-progress', 'ready', 'deprecated', 'deleted', 'in-active']
ACCOUNT_ENVIRONMENTS = ['test', 'prod']
ACCOUNT_OWNERS = ['netflix', 'dvd', 'aws', 'third-party']


class NoteSchema(Schema):
    date = fields.Date()
    text = fields.Str(required=True)


class RoleSchema(Schema):
    policyUrl = fields.Str()
    roleName = fields.Str()
    id = fields.Str(required=True)
    secondaryApprover = fields.Str(default=None, missing=None)
    googleGroup = fields.Str(required=True)


class AccountStatusSchema(Schema):
    region = fields.Str(required=True)
    status = fields.Str(validate=OneOf(ACCOUNT_STATUSES), missing='created')
    notes = fields.Nested(NoteSchema, many=True, missing=[])


class ServiceStatusSchema(Schema):
    region = fields.Str(required=True)
    enabled = fields.Boolean(missing=False)
    notes = fields.Nested(NoteSchema, many=True, missing=[])


class ServiceSchema(Schema):
    name = fields.Str(required=True)
    status = fields.Nested(ServiceStatusSchema, many=True, required=True)
    roles = fields.Nested(RoleSchema, many=True, missing=[])
    metadata = fields.Dict(missing={})


class AccountSchema(Schema):
    schemaVersion = fields.Str(missing='2')
    id = fields.Str(required=True)
    name = fields.Str(required=True)
    contacts = fields.List(fields.Email(), required=True, missing=[])
    provider = fields.Str(validate=OneOf(PROVIDERS), missing='aws')
    type = fields.Str(validate=OneOf(ACCOUNT_TYPES), missing='service')
    tags = fields.List(fields.Str(), missing=[])
    status = fields.Nested(AccountStatusSchema, many=True, missing=[])
    email = fields.Email(required=True)
    environment = fields.Str(validate=OneOf(ACCOUNT_ENVIRONMENTS), missing='prod')
    services = fields.Nested(ServiceSchema, many=True, missing=[])
    sensitive = fields.Bool(missing=False)
    description = fields.Str(required=True)
    owner = fields.Str(validate=OneOf(ACCOUNT_OWNERS), required=True, missing='netflix')
    aliases = fields.List(fields.Str(), missing=[])





"""
.. module:: swag_client.schemas.v1
    :platform: Unix
.. author:: Kevin Glisson (kglisson@netflix.com)
.. author:: Mike Grima (mgrima@netflix.com)
"""
from marshmallow import Schema, fields, validates_schema
from marshmallow.validate import Length, OneOf

from swag_client.schemas.validators import validate_fqdn


class AWSAccountSchema(Schema):
    account_number = fields.String(required=True, validate=Length(max=12, min=12))  # equal keyword not yet available in mainline marshmallow
    cloudtrail_index = fields.String(required=True)
    cloudtrail_kibana_url = fields.String()
    s3_name = fields.String()
    email = fields.Email(required=True)


class GoogleProjectSchema(Schema):
    project_id = fields.String(required=True, validate=Length(max=30))
    project_number = fields.Integer(required=True)
    project_name = fields.String(required=True)


TYPES = {'aws': AWSAccountSchema, 'gcp': GoogleProjectSchema}


# Here are define top level fields that every account would need
class AccountSchema(Schema):
    id = fields.String(required=True)
    name = fields.String(required=True)
    type = fields.String(required=True, validate=OneOf(TYPES.keys()))
    metadata = fields.Dict(required=True)
    tags = fields.List(fields.String())
    services = fields.Dict()
    cmc_required = fields.Boolean(required=True)
    description = fields.String(required=True)
    owners = fields.List(fields.Email(required=True), required=True)
    alias = fields.List(fields.String(), required=True)
    bastion = fields.String(validate=validate_fqdn)

    # Set to True if your company owns the account.
    # Set to False if this is a partner account your apps may need to know about.
    ours = fields.Boolean(required=True)

    schema_version = fields.Integer(required=True, missing='v1')

    @validates_schema
    def validate_metadata(self, data):
        TYPES[data['type']](many=True, strict=True).load([data['metadata']])
